﻿var moment = require('moment');
var cp = require('child_process');
var _ = require('lodash');
var http = require('http');
var fs = require('fs');

function sendValue(io, logger, dataType, data) {
    //logger.debug("systemStatusManager.sendValue: " + dataType + " = " + JSON.stringify(data));
    io.sockets.emit(dataType, data);
}

//---------------------------------------------
// System Status Manager
//#region System Status Manager
// Initialize the System Status Manager
exports.InitializeSystemStatusManager = function (db, logger,callback) {
    // Try getting the ErrorLog table used for tracking system status
    db.get("SELECT name FROM sqlite_master WHERE type='table' AND name='ErrorLog'", function (err, rows) {
        if (err) {
            throw logger.error('System Status - Initialize Error (Select)=' + JSON.stringify(err));
        }
        else if (rows === undefined) {
            // We read nothing, so let's create the table
            logger.info("System Status - Initialize by creating database table ...");
            db.exec("CREATE TABLE if not exists ErrorLog (ErrorMessage TEXT, ErrorType TEXT, TimeChanged TEXT )", function (err) {
                if (err !== null) {
                    logger.error('System Status - Initialize Error (Create)=' + JSON.stringify(err));
                }
                else {
                    var timeStamp = new Date();// Record the create time as the last time the I/O point value was changed
                    var timeChanged = timeStamp.toISOString();
                    
                    db.run("BEGIN")
                    db.run("INSERT INTO ErrorLog (ErrorMessage, ErrorType, TimeChanged) VALUES ('Started ErrorLog','Information','" + timeChanged + "')", function (err) {
                        if (err !== null) {
                            logger.error('System Status - Inititalize Error (Insert)=' + JSON.stringify(err));
                        }
                    });
                    db.run("COMMIT");
                }
                logger.info('System Status - Initialized new database table');
                // System Status table created
                if (typeof callback == "function") {
                    logger.info('System Status - Initialized new System Status database table');
                    callback();
                }
            });
        } else {
            // System Status table already exists
            if (typeof callback == "function") {
                logger.info('System Status - Validated System Status database table');
                callback();
            }        
        }
    });
}

// Return the current state of the system
var ErrorStatus = 'FAILURE';
exports.GetCurrentSystemStatus = function (logger) {
    logger.info("System Status Manager - Get Current System Status=" + ErrorStatus);
    return ErrorStatus;
}

//	Log a system error, warning or information message
exports.LogSystemStatus = function (eventMessage, eventType, fs, io, logger, db) {
    // Serialize the filling of the table
    var timeStamp = new Date();// Record the create time as the last time the I/O point value was changed
    var timeChanged = timeStamp.toISOString();
    
    db.run("BEGIN");
    db.run("INSERT INTO ErrorLog (ErrorMessage, ErrorType, TimeChanged) VALUES ('" + eventMessage + "','" + eventType + "','" + timeChanged + "')", function (err) {
        if (err !== null) {
            logger.error('System Status - Log Error (Insert)=' + JSON.stringify(err));
        }
    });
    db.run("COMMIT");
    
    // Depending on the type of errorType, set the LED accordingly
    ErrorStatus = eventType;
    sendValue(io, logger, "SystemStatus", { "ErrorMessage": eventMessage, "ErrorType": eventType, "TimeChanged": moment(timeChanged).format("HH:mm:ss.SSS - MMMM DD, YYYY") });
    switch (eventType) {
        case "Failure":
        case "FAILURE":
            this.FailureOperation(fs, io, logger);
            break;
        case "Warning":
        case "WARNING":
            this.WarningOperation(fs, io, logger);
            break;
        default:
            this.NormalOperation(fs, io, logger);
            break;
    }
    logger.info('System Status - Event="' + eventMessage + '" (' + eventType + ')');
}
//#endregion System Status Manager

//---------------------------------------------
// LED Manager
//#region LED Manager
// Define the configuraton for the I/O
var ledGeneralLocation = '/sys/class/gpio/gpio53';
var ledRedLocation = '/sys/class/leds/led:red';
var ledGreenLocation = '/sys/class/leds/led:green';
var ledBlueLocation = '/sys/class/leds/led:blue';

var moment = require('moment');

// Set the system LED to Normal Operation
exports.OffOperation = function (fs, io, logger) {
    // Requires for the I/O
    fs.writeFile(ledGeneralLocation + "/direction", "out", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledGeneralLocation + "/value", "1", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledRedLocation + "/trigger", "none", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledRedLocation + "/brightness", "0", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledGreenLocation + "/trigger", "none", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledGreenLocation + "/brightness", "0", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledBlueLocation + "/trigger", "none", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledBlueLocation + "/brightness", "0", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    //logger.info('System Status - NORMAL')
}

// Set the system LED to Normal Operation
exports.NormalOperation = function (fs, io, logger) {
    // Requires for the I/O
    fs.writeFile(ledGeneralLocation + "/direction", "out", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)="+JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledGeneralLocation + "/value", "1", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledRedLocation + "/trigger", "none", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledRedLocation + "/brightness", "0", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledGreenLocation + "/trigger", "none", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledGreenLocation + "/brightness", "0", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledBlueLocation + "/trigger", "none", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledBlueLocation + "/brightness", "1", function (err) {
        if (err) {
            logger.error("System Status - Normal Error (LED)=" + JSON.stringify(err));
        }
    });
    //logger.info('System Status - NORMAL')
}

// Set the system LED to Failed Operation
exports.FailureOperation = function (fs, io, logger) {
    fs.writeFile(ledRedLocation + "/brightness", "1", function (err) {
        if (err) {
            logger.error("System Status - Failure Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledGreenLocation + "/brightness", "0", function (err) {
        if (err) {
            logger.error("System Status - Failure Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledBlueLocation + "/brightness", "0", function (err) {
        if (err) {
            logger.error("System Status - Failure Error (LED)=" + JSON.stringify(err));
        }
    });
    //logger.info('System Status - FAILURE')
}

// Set the system LED to Warning Operation
exports.WarningOperation = function (fs, io, logger) {
    fs.writeFile(ledRedLocation + "/brightness", "1", function (err) {
        if (err) {
            logger.error("System Status - Warning Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledGreenLocation + "/brightness", "1", function (err) {
        if (err) {
            logger.error("System Status - Warning Error (LED)=" + JSON.stringify(err));
        }
    });
    
    fs.writeFile(ledBlueLocation + "/brightness", "0", function (err) {
        if (err) {
            logger.error("System Status - Warning Error (LED)=" + JSON.stringify(err));
        }
    });
    //logger.info('System Status - WARNING')
}
//#endregion LED Manager

//---------------------------------------------
// Crest software
//#region Crest software

// Ensure that the interfaces file is there for factory reset
exports.EnsureFactoryResetInterfacesFile = function (logger) {
    
    // If the factory interfaces file doesn't exist, then copy the existing interfaces file to the factory reset location
    //  - This would be the case if software was loaded before January 2017, 
    //    when we specifically didn't load a MACID' ed version of the interfaces file in the factory directory
    if (fs.existsSync('factory/interfaces')) {
        logger.warn('EnsureFactoryResetInterfacesFile - interfaces file already exists in the factory directory, no copy required');
    }
    else {
        const exec = cp.exec;

        exec("cp /etc/network/interfaces factory/interfaces", function (error, stdout, stderr) {
            logger.info('CopyFactoryResetInterfacesFile - Copying /etc/network/interfaces factory/interfaces');
            logger.info('stdout: ' + stdout);
            
            exec("sync", function (error, stdout, stderr) {
                logger.info('sync stdout: ' + stdout);
                //logger.error('sync stderr: ' + stderr);
                if (error !== null) {
                    logger.error('sync exec error: ' + error);
                }
                
                // Call back to finish intefaces file check
                logger.warn('EnsureFactoryResetInterfacesFile - Successfully copied interfaces file to the factory directory');
            });
        });
    }
}

// Get latest Crest version on server
exports.GetLatestCrestVersionFromServer = function (logger, callback) {
    var destination = 'newpackage.json';
    var url = 'http://www.tascsystems.com/downloads/summit/package.json';
    var file = fs.createWriteStream(destination);

    logger.debug('GetLatestAppVersionFromServer - Retrieving latest update info from "' + url + '"');
    var request = http.get(url, function (response) {
        // Check for a server response error
        logger.debug('GetLatestAppVersionFromServer - Response from server=' + response.statusCode);
        if (response.statusCode !== 200) {
            logger.error('GetLatestAppVersionFromServer - Response error=' + response.statusCode);
            callback('Error retrieving update info')
            return;
        } else {
            // Save the file
            response.pipe(file);
            file.on('finish', function () {
                logger.info('GetLatestAppVersionFromServer - Successfully retrieved "' + destination + '"');
                file.close(function () {
                    // Now that we have the file, let's get the version info
                    fs.readFile(destination, function (error, data) {
                        if (error) {
                            logger.error('GetLatestAppVersionFromServer - Parsing error=' + error);
                            callback('Error with file parse');
                            return;
                        }
                        
                        // Send back file information
                        var newPackageJson = JSON.parse(data);
                        logger.info("GetLatestAppVersionFromServer - Version=" + newPackageJson.version);
                        callback(newPackageJson.version);
                        return;
                    });

                })
            })
        }
    });

    // Check for a request error
    request.on('error', function (error) {
        if (fs.existsSync(destination)) fs.unlinkSync(destination);// Delete the file sync
        if (callback) {
            logger.error('GetLatestAppVersionFromServer - Request error=' + error);
            callback('Error retrieving latest version info via Internet!');
            return;
        }
    });

    // Check for file processing error
    file.on('error', function (err) {
        if (fs.existsSync(destination)) fs.unlinkSync(destination); // Delete the file sync
        if (callback) {
            logger.error('GetLatestAppVersionFromServer - File error=' + error.message);
            callback('Error with file');
            return;
        }
    });
}

exports.DownloadCrest = function (CrestUrl, logger, callback) {
    logger.debug('DownloadCrest - Data=' + CrestUrl);

    // Parse out the file
    var CrestUrlSplit = CrestUrl.split('/');
    var destination = CrestUrlSplit[CrestUrlSplit.length-1];
    
    // Save the file to the Crest directory 
    var file = fs.createWriteStream(destination);
    var request = http.get(CrestUrl, function (response) {
        // Check for a server response error
        logger.debug('DownloadCrest - Response from server=' + response.statusCode);
        if (response.statusCode !== 200) {
            logger.error('DownloadCrest - Response error=' + response.statusCode);
            callback('Error retrieving update info')
        } else {
            // Save the file
            response.pipe(file);
            file.on('finish', function () {
                logger.info('DownloadCrest - Successfully retrieved "' + destination + '" from URL="' + CrestUrl + '"');
                file.close(function () {
                    // New version has successfully been downloaded
                    logger.info('DownloadCrest - Successfully saved "' + destination + '"');
                    const exec = cp.exec;
                    
                    // Backup the package.json file so when we start up again, we'll know the last revision
                    var pjson = require('./package.json');
                    exec("cp package.json previous.json", function (error, stdout, stderr) {
                        logger.info('DownloadCrest - Copying package.json to previous.json (' + pjson.version + ') to remember old version');
                        logger.info('stdout: ' + stdout);
                        
                        // Backup the database
                        exec("cp data/summit.db data/summit.db."+pjson.version, function (error, stdout, stderr) {
                            logger.info('DownloadCrest - Copying data/summit.db to data/summit.db.' + pjson.version + '');
                            logger.info('stdout: ' + stdout);

                            // OK, backup of package.json and summit.db is complete so now install the new version
                            const spawn = cp.spawn;
                            logger.info('DownloadCrest - Unarchiving "' + destination + '"');
                            var unarchive = spawn("tar", ["-xvf", destination]);

                            unarchive.stdout.on('data', function (data) {
                                logger.info('stdout: ' + data);
                            });

                            unarchive.stderr.on('data', function (data) {
                                logger.error('stderr: ' + data);
                            });

                            unarchive.on('close', function (code) {
                                logger.info('unarchive process exited with code ' + code);
                                // Ensure that the file contents are actually written to disk
                                //  Without this command, the file contents may still be in the cache
                                //  and when rebooted, might be lost
                                exec("sync", function (error, stdout, stderr) {
                                    logger.info('sync stdout: ' + stdout);
                                    //logger.error('sync stderr: ' + stderr);
                                    if (error !== null) {
                                        logger.error('sync exec error: ' + error);
                                    }

                                    // Call back to finish up our configuration changes
                                    //  return back to calling API function
                                    logger.info('DownloadCrest - Finished unarchiving "' + destination + '"');
                                    callback(destination);
                                });
                            });
                        });
                    });
                });
            });
        }
    });
    
    // Check for a request error
    request.on('error', function (error) {
        fs.unlink(destination);// Delete the file async
        if (callback) {
            logger.error('DownloadCrest - Request error=' + error);
            return callback('Error with file request');
        }
    });
    
    // Check for file processing error
    file.on('error', function (err) {
        fs.unlink(destination); // Delete the file async
        if (callback) {
            logger.error('DownloadCrest - File error=' + error.message);
            return callback('Error with file');
        }
    });

}
//#endregion Crest software

